%------------------------------------------------------------------------------%
%------------------------------------------------------------------------------%
%
% Module: battle
% Author: Ralph Becket <rwab1@cam.sri.com>, from shell by Peter Ross
%
% An engine to play the game battleships.  The players of the game must
% obey the constraints imposed by the typeclass player(T).
%
%------------------------------------------------------------------------------%
%------------------------------------------------------------------------------%

:- module battle.

:- interface.

% :- import_module io.
:- use_module io.

%------------------------------------------------------------------------------%

	% An (x,y) coordinate, where (0,0) is the top left corner of
	% the board, and the x coordinates run in the across direction.
:- type locn 
	--->	locn(int, int).

	% Different types of ships.
:- type ship
	--->	battleship
	;	cruiser
	;	destroyer.

	% Result of the guess.
:- type result
	--->	hit(locn)
	;	sink(locn, ship)
	;	miss(locn)
	;	fleet_destroyed.

	% Direction the ship is pointing.
:- type dirn
	--->	across
	;	down.

	% The player typeclass, your players must be an instance of this
	% typeclass.
:- typeclass player(T) where [
		pred place_ship(locn, ship, locn, dirn, T, T,
				io__state, io__state),
		mode place_ship(in, in, out, out, in, out, di, uo) is det,

		pred guess(locn, T, T, io__state, io__state),
		mode guess(out, in, out, di, uo) is det,

		pred your_shot_result(battle__result, T, T,
				io__state, io__state),
		mode your_shot_result(in, in, out, di, uo) is det,

		pred their_shot_result(battle__result, T, T,
				io__state, io__state),
		mode their_shot_result(in, in, out, di, uo) is det
	].

	% The game engine.
:- pred play_battleships(A, B, io__state, io__state) <= (player(A), player(B)).
:- mode play_battleships(in, in, di, uo) is det.

	% How big is each size of ship.
:- func ship_size(ship) = int.

%------------------------------------------------------------------------------%
%------------------------------------------------------------------------------%

:- implementation.

:- import_module std_util, list, bool, int.

% ---------------------------------------------------------------------------- %

play_battleships(A0, B0) -->
	{ new_board(BoardA0) },
	set_up_player(A0, BoardA0, A, BoardA),
	{ new_board(BoardB0) },
	set_up_player(B0, BoardB0, B, BoardB),
	main_loop(A, BoardA, B, BoardB).

% ---------------------------------------------------------------------------- %

:- pred main_loop(T1, board, T2, board, io__state, io__state)
		<= (player(T1), player(T2)).
:- mode main_loop(in, in, in, in, di, uo) is det.

main_loop(A0, BoardA, B0, BoardB0) -->
	take_turn(A0, B0, BoardB0, A, B, BoardB, AllDone),
	(if { AllDone = no } then
		main_loop(B, BoardB, A, BoardA)
	else
		{ true }
	).

% ---------------------------------------------------------------------------- %

:- pred take_turn(T1, T2, board, T1, T2, board, bool, io__state, io__state)
		<= (player(T1), player(T2)).
:- mode take_turn(in, in, in, out, out, out, out, di, uo) is det.

take_turn(A0, B0, BoardB0, A, B, BoardB, AllDone) -->
	guess(Locn, A0, A1),
	{ update_board(BoardB0, Locn, BoardB, Result) },
	your_shot_result(Result, A1, A),
	their_shot_result(Result, B0, B),
	{ Result = fleet_destroyed -> AllDone = yes ; AllDone = no }.

% ---------------------------------------------------------------------------- %

:- pred set_up_player(T, board, T, board, io__state, io__state) <= player(T).
:- mode set_up_player(in, in, out, out, di, uo) is det.

set_up_player(P0, BoardP0, P, BoardP) -->
	set_up_ship(destroyer,  P0, BoardP0, P1, BoardP1),
	set_up_ship(destroyer,  P1, BoardP1, P2, BoardP2),
	set_up_ship(destroyer,  P2, BoardP2, P3, BoardP3),
	set_up_ship(cruiser,    P3, BoardP3, P4, BoardP4),
	set_up_ship(cruiser,    P4, BoardP4, P5, BoardP5),
	set_up_ship(battleship, P5, BoardP5, P , BoardP ).

% ---------------------------------------------------------------------------- %

:- pred set_up_ship(ship, T, board, T, board, io__state, io__state) <=player(T).
:- mode set_up_ship(in, in, in, out, out, di, uo) is det.

set_up_ship(Ship, P0, BoardP0, P, BoardP) -->
	place_ship(board_size, Ship, Locn, Dirn, P0, P),
	{ ship_locations(Ship, Locn, Dirn, Locns) },
	{ add_ship(BoardP0, Ship, Locns, BoardP) }.

% ---------------------------------------------------------------------------- %

:- pred ship_locations(ship, locn, dirn, list(locn)).
:- mode ship_locations(in, in, in, out) is det.

ship_locations(destroyer,   locn(X, Y), across,
		[locn(X, Y), locn(X+1, Y)]).
ship_locations(destroyer,  locn(X, Y), down,
		[locn(X, Y), locn(X, Y+1)]).

ship_locations(cruiser,     locn(X, Y), across,
		[locn(X, Y), locn(X+1, Y), locn(X+2, Y)]).
ship_locations(cruiser,    locn(X, Y), down,
		[locn(X, Y), locn(X, Y+1), locn(X, Y+2)]).

ship_locations(battleship, locn(X, Y), across,
		[locn(X, Y), locn(X+1, Y), locn(X+2, Y), locn(X+3, Y)]).
ship_locations(battleship, locn(X, Y), down,
		[locn(X, Y), locn(X, Y+1), locn(X, Y+2), locn(X, Y+3)]).

%------------------------------------------------------------------------------%

ship_size(battleship) = 4.
ship_size(cruiser) = 3.
ship_size(destroyer) = 2.

% ---------------------------------------------------------------------------- %
% ---------------------------------------------------------------------------- %

:- type board == list(ship_placement).

:- type ship_placement == pair(ship, list(locn)).

% ---------------------------------------------------------------------------- %

:- func board_size = locn.
:- mode board_size = out is det.

board_size = locn(10,10).

% ---------------------------------------------------------------------------- %

:- pred new_board(board).
:- mode new_board(out) is det.

new_board([]).

% ---------------------------------------------------------------------------- %

	% XXX should really check for overlaps and do bounds checking here...

:- pred add_ship(board, ship, list(locn), board).
:- mode add_ship(in, in, in, out) is det.

add_ship(Board, Ship, Locns, [(Ship-Locns) | Board]).

% ---------------------------------------------------------------------------- %

	% XXX should really do bounds checking here...

:- pred update_board(board, locn, board, result).
:- mode update_board(in, in, out, out) is det.

update_board(Board0, Locn, Board, Result) :-
	remove_victim(Board0, Locn, Victim0, Board1) ->
		update_victim(Victim0, Locn, Victim),
		add_updated_victim(Victim, Board1, Locn, Board, Result)
	;
		Board = Board0,
		Result = miss(Locn).

% ---------------------------------------------------------------------------- %

:- pred remove_victim(board, locn, ship_placement, board).
:- mode remove_victim(in, in, out, out) is semidet.

remove_victim([Ship-Placement | Board0], Locn, Victim, Board) :-
	list__member(Locn, Placement) ->
		Victim = (Ship-Placement),
		Board = Board0
	;
		Board = [Ship-Placement | Board1],
		remove_victim(Board0, Locn, Victim, Board1).

% ---------------------------------------------------------------------------- %

:- pred update_victim(ship_placement, locn, ship_placement).
:- mode update_victim(in, in, out) is det.

update_victim(Ship-Placement0, Locn, Ship-Placement) :-
	list__delete_all(Placement0, Locn, Placement).

% ---------------------------------------------------------------------------- %

:- pred add_updated_victim(ship_placement, board, locn, board, result).
:- mode add_updated_victim(in, in, in, out, out) is det.

add_updated_victim(Ship-Placement, Board0, Locn, Board, Result) :-
	(
			% That was the last ship!
		Placement = [],
		Board0 = []
	->
		Board = [],
		Result = fleet_destroyed
	;
			% This ship has been sunk - remove it from the board.
		Placement = []
	->
		Board = Board0,
		Result = sink(Locn, Ship)
	;
			% A non-fatal hit.
		Board = [Ship-Placement | Board0],
		Result = hit(Locn)
	).

%------------------------------------------------------------------------------%
%------------------------------------------------------------------------------%
