% ---------------------------------------------------------------------------- %
% ---------------------------------------------------------------------------- %
%
% Module: rwab1__logging
% Author: Ralph Becket <rwab1@cam.sri.com>
%
% OUTLINE
%
%   This module implements the player typeclass as a wrapper for another player
%   object.  All actions of the player object are logged to a file specified
%   when the logging object is created.
%
% NOTES
%
% * The logging output is simple but effective.
%
% * A blank line is inserted before each `guess' to aid legibility.
%
% * Output is to a stream.  The stream is flushed after each output so the
%   log can be viewed as it is added to.  The stream is closed after either
%   fleet is destroyed.  Any further operation on the enclosed object should
%   therefore lead to an IO error, since the stream is only opened when the
%   logging object is created.
%
% ---------------------------------------------------------------------------- %
% ---------------------------------------------------------------------------- %



:- module rwab1__logging.

:- interface.

:- import_module io.

:- type logging(T).

:- instance player(logging(T)) <= player(T) where [
		pred(place_ship/8) is logging__place_ship,
		pred(guess/5) is logging__guess,
		pred(your_shot_result/5) is logging__your_shot_result,
		pred(their_shot_result/5) is logging__their_shot_result
	].

	% Returns the initial state for use by the logging player.
:- pred logging__logging(T, string, logging(T), io__state, io__state) <= player(T).
:- mode logging__logging(in, in, out, di, uo) is det.

:- pred logging__place_ship(locn, ship, locn, dirn, logging(T), logging(T), io__state, io__state) <= player(T).
:- mode logging__place_ship(in, in, out, out, in, out, di, uo) is det.

:- pred logging__guess(locn, logging(T), logging(T), io__state, io__state) <= player(T).
:- mode logging__guess(out, in, out, di, uo) is det.

:- pred logging__your_shot_result(battle__result, logging(T), logging(T), io__state, io__state) <= player(T).
:- mode logging__your_shot_result(in, in, out, di, uo) is det.

:- pred logging__their_shot_result(battle__result, logging(T), logging(T), io__state, io__state) <= player(T).
:- mode logging__their_shot_result(in, in, out, di, uo) is det.

% ---------------------------------------------------------------------------- %
% ---------------------------------------------------------------------------- %

:- implementation.

:- import_module require.

:- type logging(T) 
	--->	logging(T, io__output_stream).

% ---------------------------------------------------------------------------- %

logging__logging(Player, FileName, logging(Player, Stream)) -->
	io__open_output(FileName, IO_S),
	(
		{ IO_S = ok(Stream0) },
		{ Stream = Stream0 }
	;
		{ IO_S = error(ErrorCode) },
		{ io__error_message(ErrorCode, ErrorMessage) },
		{ error(ErrorMessage) }
	).

% ---------------------------------------------------------------------------- %

logging__place_ship(BoardSize, Ship, Locn, Dirn, Self0, Self) -->
	{ Self0 = logging(Player0, Stream) },
	place_ship(BoardSize, Ship, Locn, Dirn, Player0, Player),
	io__print(Stream, "ship placement "),
	io__print(Stream, Ship), io__print(Stream, " "),
	io__print(Stream, Locn), io__print(Stream, " "),
	io__print(Stream, Dirn),
	io__nl(Stream),
	io__flush_output(Stream),
	{ Self = logging(Player, Stream) }.

% ---------------------------------------------------------------------------- %

logging__guess(Locn, Self0, Self) -->
	{ Self0 = logging(Player0, Stream) },
	guess(Locn, Player0, Player),
	io__nl(Stream),
	io__print(Stream, "guess "),
	io__print(Stream, Locn),
	io__nl(Stream),
	io__flush_output(Stream),
	{ Self = logging(Player, Stream) }.

% ---------------------------------------------------------------------------- %

logging__your_shot_result(Result, Self0, Self) -->
	{ Self0 = logging(Player0, Stream) },
	your_shot_result(Result, Player0, Player),
	io__print(Stream, "result "),
	io__print(Stream, Result),
	io__nl(Stream),
	io__flush_output(Stream),
	{ Self = logging(Player, Stream) }.

% ---------------------------------------------------------------------------- %

logging__their_shot_result(Result, Self0, Self) -->
	{ Self0 = logging(Player0, Stream) },
	their_shot_result(Result, Player0, Player),
	io__print(Stream, "enemy "),
	io__print(Stream, Result),
	io__nl(Stream),
	io__flush_output(Stream),
	( { Result = fleet_destroyed } ->
		io__close_output(Stream)
	;
		{ true }
	),
	{ Self = logging(Player, Stream) }.

