% ---------------------------------------------------------------------------- %
% ---------------------------------------------------------------------------- %
%
% Module: rwab1__robot
% Author: Ralph Becket <rwab1@cam.sri.com>
%
% ALGORITHM
%
%   Robot battleships player.  The robot places its ships at random (well,
%   insofar as the random number generator allows -- it isn't currently that
%   good -- and the fact that the random seed is always the same...)  It then
%   takes pot-shots on the grid (X + Y) mod 2 = 0, which is guaranteed to
%   eventually hit any ship since each ship must take up at least two adjacent
%   locations.  Once an enemy ship has been found, the robot starts searching
%   outwards along the cardinal directions.  The second hit tells the robot
%   which *line* the ship must be on and so the robot can further restrict the
%   locus of points to search.  Similarly, a miss on a line restricts the
%   search to a semi-line.
%
% POSSIBLE IMPROVEMENTS
%
% * The robot's `random' search is O(n^2) for the nth shot.  This could be
%   reduced to O(1) by adding some extra state to the robot.
%
% * The robot is somewhat taciturn.
%
% * On a small enough grid (i.e. unamusingly small), the `random' placement
%   algorithm may not terminate (since placing ship n may prohibit the
%   placement of ship n+1).
%
% * It's not clear whether the robot should produce any output.  Bit silly
%   when playing against a human...
%
% ---------------------------------------------------------------------------- %
% ---------------------------------------------------------------------------- %



:- module rwab1__robot.

:- interface.

:- import_module io.

:- type robot.

:- instance player(robot) where [
		pred(place_ship/8)		is robot__place_ship,
		pred(guess/5)			is robot__guess,
		pred(your_shot_result/5)	is robot__your_shot_result,
		pred(their_shot_result/5)	is robot__their_shot_result
	].

	% Returns the initial state for use by the robot player.
:- pred robot__robot(robot::out) is det.

:- pred robot__place_ship(locn, ship, locn, dirn, robot, robot, io__state, io__state).
:- mode robot__place_ship(in, in, out, out, in, out, di, uo) is det.

:- pred robot__guess(locn, robot, robot, io__state, io__state).
:- mode robot__guess(out, in, out, di, uo) is det.

:- pred robot__your_shot_result(battle__result, robot, robot, io__state, io__state).
:- mode robot__your_shot_result(in, in, out, di, uo) is det.

:- pred robot__their_shot_result(battle__result, robot, robot, io__state, io__state).
:- mode robot__their_shot_result(in, in, out, di, uo) is det.

% ---------------------------------------------------------------------------- %
% ---------------------------------------------------------------------------- %

:- implementation.

:- import_module list, string, int, std_util, random, require.
:- import_module rwab1__view.

:- type robot
	--->	robot(
			robot_state,	% Mode of operation.
			view		% State of play.
		).

:- type robot_state
	--->	searching_at_random
	;	searching_in_plane(nsew_locns)
	;	searching_in_line(nsew_locns).

:- type nsew_locns == list(nsew_locn).

:- type nsew_locn == pair(nsew, locn).

:- type nsew
	--->	n
	;	s
	;	e
	;	w.

% ---------------------------------------------------------------------------- %

robot__robot(robot(searching_at_random, View)) :-
	view__init(View).

% ---------------------------------------------------------------------------- %

robot__place_ship(BoardSize, Ship, Locn, Dirn, robot(Mode, View0), robot(Mode, View), IO, IO) :-
	random_placement(BoardSize, Ship, Locn, Dirn, View0, View).

% ---------------------------------------------------------------------------- %
% This just hunts until it finds a place to put a local ship.  Not
% very efficient, but who cares.

:- pred random_placement(locn, ship, locn, dirn, view, view).
:- mode random_placement(in, in, out, out, in, out) is det.

random_placement(BoardSize, Ship, Locn, Dirn, View0, View) :-
	random__init(0, S),
	random_placement_0(S, BoardSize, Ship, Locn, Dirn, View0, View).

% ---------------------------------------------------------------------------- %

:- pred random_placement_0(random__supply, locn, ship, locn, dirn, view, view).
:- mode random_placement_0(mdi, in, in, out, out, in, out) is det.

random_placement_0(S0, BoardSize, Ship, Locn, Dirn, View0, View) :-

	BoardSize = locn(Cols, Rows),

	random__random(Y, S0, S1),
	random__random(X, S1, S2),
	Locn0 = locn(X mod Cols, Y mod Rows),

	random__random(D, S2, S),
	( D /\ 1 = 0 -> Dirn0 = across ; Dirn0 = down ),

	view__place_local_ship(BoardSize,Ship,Locn0,Dirn0,Result,View0,View1),

	(
		Result = ok,
		Locn = Locn0,
		Dirn = Dirn0,
		View = View1
	;
		Result = error(_),
		random_placement_0(S, BoardSize, Ship, Locn, Dirn, View1, View)
	).

% ---------------------------------------------------------------------------- %

robot__guess(Locn, robot(Mode, View), robot(Mode, View), IO, IO) :-

	(
		Mode = searching_at_random,
		random_guess(View, Locn)
	;
		Mode = searching_in_plane(NSEWLocns),
		try_potential_site(NSEWLocns, Locn)
	;
		Mode = searching_in_line(NSEWLocns),
		try_potential_site(NSEWLocns, Locn)
	).

% ---------------------------------------------------------------------------- %

:- pred random_guess(view, locn).
:- mode random_guess(in, out) is det.

random_guess(View, Locn) :-

	random_guess_0(0, View, Locn).

% ---------------------------------------------------------------------------- %

:- pred random_guess_0(int, view, locn).
:- mode random_guess_0(in, in, out) is det.

random_guess_0(N, View, Locn) :-

	view__board_size(View, locn(Cols, Rows)),

	Y = (N // Cols) mod Rows,
	X = (N + (Y /\ 1) - (Cols /\ 1)) mod Cols,
	Locn0 = locn(X, Y),

	( view__object_at_enemy_locn(View, Locn0, sea) ->
		Locn = Locn0
	;
		random_guess_0(N+2, View, Locn)
	).

% ---------------------------------------------------------------------------- %

:- pred try_potential_site(nsew_locns, locn).
:- mode try_potential_site(in, out) is det.

try_potential_site([], _) :-
	error("rwab1__robot__search/2: wagga wagga.").

try_potential_site([_NSEW-Locn | _], Locn).

% ---------------------------------------------------------------------------- %

robot__your_shot_result(Result, robot(Mode0, View0), robot(Mode, View)) -->

	{ update_mode(Mode0, Result, View0, Mode) },

	(
		{ Result = fleet_destroyed },
		{ View = View0 }
	;
		{ Result = miss(Locn) },
		{ update_enemy_view(Locn, miss, View0, View) }
	;
		{ Result = sink(Locn, _Ship) },
		{ update_enemy_view(Locn, hit, View0, View) }
	;
		{ Result = hit(Locn) },
		{ update_enemy_view(Locn, hit, View0, View) }
	).

% ---------------------------------------------------------------------------- %

:- pred update_mode(robot_state, battle__result, view, robot_state).
:- mode update_mode(in, in, in, out) is det.

update_mode(
	searching_at_random,
	miss(_Locn),
	_View,
	searching_at_random
).

update_mode(
	searching_at_random,
	hit(locn(X, Y)),
	View,
	searching_in_plane(Plane)
) :-
		% Set of candidate locations for rest of ship.
	Plane0 = [
		n-locn(X, Y+1), s-locn(X, Y-1), e-locn(X+1, Y), w-locn(X-1, Y),
		n-locn(X, Y+2), s-locn(X, Y-2), e-locn(X+2, Y), w-locn(X-2, Y),
		n-locn(X, Y+3), s-locn(X, Y-3), e-locn(X+3, Y), w-locn(X-3, Y)
	],
		% Remove locations not on the board.
	list__filter(
		( pred((_NSEW-Locn)::in) is semidet :-
			view__locn_is_on_board(Locn, View)
		),
		Plane0,
		Plane
	).

update_mode(
	searching_at_random,
	sink(_, _),
	_View,
	searching_at_random
).

update_mode(
	searching_at_random,
	fleet_destroyed,
	_View,
	searching_at_random
).



update_mode(
	searching_in_plane(NSEWLocns0),
	miss(_Locn), 
	_View,
	searching_in_plane(NSEWLocns)
) :-
	purge_head_nsew(NSEWLocns0, NSEWLocns).

update_mode(
	searching_in_plane(NSEWLocns0),
	hit(_Locn), 
	_View,
	searching_in_line(NSEWLocns)
) :-
	purge_head_perpendicular_nsew(NSEWLocns0, NSEWLocns).

update_mode(
	searching_in_plane(_),
	sink(_Locn, _Ship), 
	_View,
	searching_at_random
).

update_mode(
	searching_in_plane(_),
	fleet_destroyed,
	_View,
	searching_at_random
).



update_mode(
	searching_in_line(NSEWLocns0),
	miss(_Locn), 
	_View,
	searching_in_line(NSEWLocns)
) :-
	purge_head_nsew(NSEWLocns0, NSEWLocns).

update_mode(
	searching_in_line(NSEWLocns0),
	hit(_Locn), 
	_View,
	searching_in_line(NSEWLocns)
) :-
	purge_head(NSEWLocns0, NSEWLocns).

update_mode(
	searching_in_line(_),
	sink(_Locn, _Ship), 
	_View,
	searching_at_random
).

update_mode(
	searching_in_line(_),
	fleet_destroyed,
	_View,
	searching_at_random
).

% ---------------------------------------------------------------------------- %

:- pred purge_nsew(nsew, nsew_locns, nsew_locns).
:- mode purge_nsew(in, in, out) is det.

purge_nsew(_NSEW, [], []).

purge_nsew(NSEW, [NSEW0-Locn0 | NSEWLocns0], NSEWLocns) :-
	NSEW = NSEW0 ->
		purge_nsew(NSEW, NSEWLocns0, NSEWLocns)
	;
		NSEWLocns = [NSEW0-Locn0 | NSEWLocns1],
		purge_nsew(NSEW, NSEWLocns0, NSEWLocns1).

% ---------------------------------------------------------------------------- %

:- pred purge_head_nsew(nsew_locns, nsew_locns).
:- mode purge_head_nsew(in, out) is det.

purge_head_nsew([], _) :-
	error("rwab1__robot__purge_head_nsew/2: pa-ting! pa-ting!").

purge_head_nsew([NSEW-_Locn | NSEWLocns0], NSEWLocns) :-
	purge_nsew(NSEW, NSEWLocns0, NSEWLocns).

% ---------------------------------------------------------------------------- %

:- pred purge_head_perpendicular_nsew(nsew_locns, nsew_locns).
:- mode purge_head_perpendicular_nsew(in, out) is det.

purge_head_perpendicular_nsew([], _) :-
	error("rwab1__robot__purge_head_perpendicular_nsew/2: pa-ting! pa-ting!").

purge_head_perpendicular_nsew([n-_Locn | NSEWLocns0], NSEWLocns) :-
	purge_nsew(e, NSEWLocns0, NSEWLocns1),
	purge_nsew(w, NSEWLocns1, NSEWLocns ).

purge_head_perpendicular_nsew([s-_Locn | NSEWLocns0], NSEWLocns) :-
	purge_nsew(e, NSEWLocns0, NSEWLocns1),
	purge_nsew(w, NSEWLocns1, NSEWLocns ).

purge_head_perpendicular_nsew([e-_Locn | NSEWLocns0], NSEWLocns) :-
	purge_nsew(n, NSEWLocns0, NSEWLocns1),
	purge_nsew(s, NSEWLocns1, NSEWLocns ).

purge_head_perpendicular_nsew([w-_Locn | NSEWLocns0], NSEWLocns) :-
	purge_nsew(n, NSEWLocns0, NSEWLocns1),
	purge_nsew(s, NSEWLocns1, NSEWLocns ).

% ---------------------------------------------------------------------------- %

:- pred purge_head(nsew_locns, nsew_locns).
:- mode purge_head(in, out) is det.

purge_head([], _) :-
	error("rwab1__robot__purge_head/2: pa-ting! pa-ting!").

purge_head([_ | NSEWLocns], NSEWLocns).

% ---------------------------------------------------------------------------- %

robot__their_shot_result(Result, robot(Mode, View0), robot(Mode, View)) -->
	(
		{ Result = fleet_destroyed },
		{ View = View0 }
	;
		{ Result = miss(Locn) },
		{ update_local_view(Locn, miss, View0, View) }
	;
		{ Result = sink(Locn, _Ship) },
		{ update_local_view(Locn, hit, View0, View) }
	;
		{ Result = hit(Locn) },
		{ update_local_view(Locn, hit, View0, View) }
	).
