% ---------------------------------------------------------------------------- %
% ---------------------------------------------------------------------------- %
%
% Module: rwab1__view
% Author: Ralph Becket <rwab1@cam.sri.com>
%
% OUTLINE
% 
%   This module provides a means of keeping track of the position of local
%   ships, the effect of enemy fire and the effect of local fire.  Means to
%   display the current board view in a human readable fashion are also
%   provided.
%
% NOTES
%
% * Some of the auxiliary predicates are special cased based upon the length
%   of the various ships as defined in the spec.  This approach has the virtue
%   of great simplicity, although a change to the spec. would mean that this
%   code would also need changing.
%
% ---------------------------------------------------------------------------- %
% ---------------------------------------------------------------------------- %



:- module rwab1__view.

:- interface.

:- import_module io.

:- type view.

:- type object
	--->	sea
	;	miss
	;	hit
	;	ship(ship).

:- type placement_result
	--->	ok
	;	error(string).		% Error message.

:- pred view__init(view).
:- mode view__init(out) is det.

:- pred view__place_local_ship(locn, ship, locn, dirn, placement_result, view, view).
:- mode view__place_local_ship(in, in, in, in, out, in, out) is det.

:- pred view__update_local_view(locn, object, view, view).
:- mode view__update_local_view(in, in, in, out) is det.

:- pred view__update_enemy_view(locn, object, view, view).
:- mode view__update_enemy_view(in, in, in, out) is det.

:- pred view__board_size(view, locn).		% locn(Cols, Rows).
:- mode view__board_size(in, out) is det.

:- pred view__locn_is_on_board(locn, view).
:- mode view__locn_is_on_board(in, in) is semidet.

:- pred view__object_at_enemy_locn(view, locn, object).
:- mode view__object_at_enemy_locn(in, in, out) is det.

:- pred view__display(view, io__state, io__state).
:- mode view__display(in, di, uo) is det.

% ---------------------------------------------------------------------- %

:- implementation.

:- import_module map, list, int, string.

:- type view
	--->	view(locn, board, board).	% BoardSize, Enemy, Local

:- type board == map(locn, object).

% ---------------------------------------------------------------------- %

view__init(view(locn(-1,-1),B,B)) :-
	map__init(B).

% ---------------------------------------------------------------------- %

view__place_local_ship(BoardSize, Ship, Locn, Dirn, Result, View0, View) :-

	( \+ within_limits(Ship, Locn, Dirn, BoardSize) ->
		View = View0,
		Result = error("* That doesn't fit on the board.")

	; \+ doesnt_collide_local(Ship, Locn, Dirn, View0) ->
		View = View0,
		Result = error("* You can't place one ship across another.")

	;
		add_local_ship(Ship, Locn, Dirn, View0, view(_,E,L)),
		View = view(BoardSize,E,L),
		Result = ok
	).

% ---------------------------------------------------------------------- %

:- pred within_limits(ship, locn, dirn, locn).
:- mode within_limits(in, in, in, in) is semidet.

within_limits(destroyer, locn(X, Y), Dirn, locn(Cols, Rows)) :-
	dirn_deltas(Dirn, DX, DY),
	0 =< X,  X + 2 * DX =< Cols,
	0 =< Y,  Y + 2 * DY =< Rows.

within_limits(cruiser, locn(X, Y), Dirn, locn(Cols, Rows)) :-
	dirn_deltas(Dirn, DX, DY),
	0 =< X,  X + 3 * DX =< Cols,
	0 =< Y,  Y + 3 * DY =< Rows.

within_limits(battleship, locn(X, Y), Dirn, locn(Cols, Rows)) :-
	dirn_deltas(Dirn, DX, DY),
	0 =< X,  X + 4 * DX =< Cols,
	0 =< Y,  Y + 4 * DY =< Rows.

% ---------------------------------------------------------------------- %

:- pred doesnt_collide_local(ship, locn, dirn, view).
:- mode doesnt_collide_local(in, in, in, in) is semidet.

doesnt_collide_local(destroyer, locn(X, Y), Dirn, view(_,_,L)) :-
	dirn_deltas(Dirn, DX, DY),
	\+ map__search(L, locn(X + 0 * DX, Y + 0 * DY), _),
	\+ map__search(L, locn(X + 1 * DX, Y + 1 * DY), _).

doesnt_collide_local(cruiser, locn(X, Y), Dirn, view(_,_,L)) :-
	dirn_deltas(Dirn, DX, DY),
	\+ map__search(L, locn(X + 0 * DX, Y + 0 * DY), _),
	\+ map__search(L, locn(X + 1 * DX, Y + 1 * DY), _),
	\+ map__search(L, locn(X + 2 * DX, Y + 2 * DY), _).

doesnt_collide_local(battleship, locn(X, Y), Dirn, view(_,_,L)) :-
	dirn_deltas(Dirn, DX, DY),
	\+ map__search(L, locn(X + 0 * DX, Y + 0 * DY), _),
	\+ map__search(L, locn(X + 1 * DX, Y + 1 * DY), _),
	\+ map__search(L, locn(X + 2 * DX, Y + 2 * DY), _),
	\+ map__search(L, locn(X + 3 * DX, Y + 3 * DY), _).

% ---------------------------------------------------------------------- %

:- pred add_local_ship(ship, locn, dirn, view, view).
:- mode add_local_ship(in, in, in, in, out) is det.

add_local_ship(destroyer, locn(X, Y), Dirn, view(BS,E,L0), view(BS,E,L)) :-
	dirn_deltas(Dirn, DX, DY),
	map__set(L0, locn(X + 0 * DX, Y + 0 * DY), ship(destroyer), L1),
	map__set(L1, locn(X + 1 * DX, Y + 1 * DY), ship(destroyer), L ).

add_local_ship(cruiser, locn(X, Y), Dirn, view(BS,E,L0), view(BS,E,L)) :-
	dirn_deltas(Dirn, DX, DY),
	map__set(L0, locn(X + 0 * DX, Y + 0 * DY), ship(cruiser), L1),
	map__set(L1, locn(X + 1 * DX, Y + 1 * DY), ship(cruiser), L2),
	map__set(L2, locn(X + 2 * DX, Y + 2 * DY), ship(cruiser), L ).

add_local_ship(battleship, locn(X, Y), Dirn, view(BS,E,L0), view(BS,E,L)) :-
	dirn_deltas(Dirn, DX, DY),
	map__set(L0, locn(X + 0 * DX, Y + 0 * DY), ship(battleship), L1),
	map__set(L1, locn(X + 1 * DX, Y + 1 * DY), ship(battleship), L2),
	map__set(L2, locn(X + 2 * DX, Y + 2 * DY), ship(battleship), L3),
	map__set(L3, locn(X + 3 * DX, Y + 3 * DY), ship(battleship), L ).

% ---------------------------------------------------------------------- %

view__update_local_view(Locn, Obj, view(BS,E,L0), view(BS,E,L)) :-
	map__set(L0, Locn, Obj, L).

% ---------------------------------------------------------------------- %

view__update_enemy_view(Locn, Obj, view(BS,E0,L), view(BS,E,L)) :-
	map__set(E0, Locn, Obj, E).

% ---------------------------------------------------------------------- %

view__board_size(view(BoardSize,_,_), BoardSize).

% ---------------------------------------------------------------------------- %

view__locn_is_on_board(locn(C,R), view(locn(Cols,Rows),_,_)) :-
	0 =< C, C < Cols,
	0 =< R, R < Rows.

% ---------------------------------------------------------------------- %

view__object_at_enemy_locn(view(_,E,_), Locn, Obj) :-
	map__search(E, Locn, Obj0) ->
		Obj = Obj0
	;
		Obj = sea.

% ---------------------------------------------------------------------- %

view__display(view(BoardSize, Enemy, Local)) -->
	{ board_to_strings(BoardSize, Enemy, EnemyStrings) },
	{ board_to_strings(BoardSize, Local, LocalStrings) },
	show_strings(["BAD GUYS" | EnemyStrings], ["GOOD GUYS" | LocalStrings]).

% ---------------------------------------------------------------------- %

:- pred board_to_strings(locn, board, list(string)).
:- mode board_to_strings(in, in, out) is det.

board_to_strings(locn(Cols,Rows), Board, [ColIndicesString | RowStrings]) :-
	column_indices_string(Cols, ColIndicesString),
	row_strings(Rows, Cols, Board, [], RowStrings).

% ---------------------------------------------------------------------- %

:- pred column_indices_string(int, string).
:- mode column_indices_string(in, out) is det.

column_indices_string(N, S) :-
	N > 0 ->
		column_indices_string(N - 1, S0),
		string__format("%s%2d", [s(S0), i(N-1)], S)
	;
		S = "   ".

% ---------------------------------------------------------------------- %

:- pred row_strings(int, int, board, list(string), list(string)).
:- mode row_strings(in, in, in, in, out) is det.

row_strings(H, W, Board, RowStrings0, RowStrings) :-
	H > 0 ->
		row_string(H, W, Board, RowString),
		row_strings(H-1, W, Board, [RowString|RowStrings0], RowStrings)
	;
		RowStrings = RowStrings0.

% ---------------------------------------------------------------------- %

:- pred row_string(int, int, board, string).
:- mode row_string(in, in, in, out) is det.

row_string(H, W, Board, RowString) :-
	W > 0 ->
		row_string(H, W-1, Board, RowString0),
		image_at_locn(locn(W-1, H-1), Board, Image),
		string__format("%s%s", [s(RowString0), s(Image)], RowString)
	;
		string__format("%2d ", [i(H-1)], RowString).

% ---------------------------------------------------------------------- %

:- pred image_at_locn(locn, board, string).
:- mode image_at_locn(in, in, out) is det.

image_at_locn(Locn, Board, Image) :-
	map__search(Board, Locn, Obj) ->
		(
			Obj = sea,	Image = " ."
		;	Obj = miss,	Image = "[]"
		;	Obj = hit,	Image = "**"
		;	Obj = ship(S),	Image = ship_image(S)
		)
	;
		Image = " .".		% Default case.

% ---------------------------------------------------------------------- %

:- func ship_image(ship) = string.
:- mode ship_image(in) = out is det.

ship_image(battleship)	= "BB".
ship_image(cruiser)	= "CC".
ship_image(destroyer)	= "DD".

% ---------------------------------------------------------------------- %

:- pred show_strings(list(string), list(string), io__state, io__state).
:- mode show_strings(in, in, di, uo) is det.

show_strings([], []) --> [].

show_strings([], [_|_]) --> [].

show_strings([_|_], []) --> [].

show_strings([E | Es], [L | Ls]) -->
	io__format("%-40s%s\n", [s(E), s(L)]),
	show_strings(Es, Ls).

% ---------------------------------------------------------------------------- %

:- pred dirn_deltas(battle__dirn, int, int).
:- mode dirn_deltas(in, out, out) is det.

dirn_deltas(across, 1, 0).
dirn_deltas(down,   0, 1).

